from typing import Callable, Any

# Function type defnitions: prefix, postfix and replace.
type PrefixFnType = Callable[..., tuple[Any, Any]]
type PostfixFnType = Callable[..., Any]
type ReplaceFnType = Callable[..., Any]


# Continue executing original & postfix
CONTINUE_EXEC = "continue"
# Continue executing original, but not postfix
CONTINUE_WITHOUT_POSTFIX = "continue_npf"
# Don't execute anything else
STOP_EXEC = "stop"


from .core import *
from .injector import *
from .hook import *


class PatchManager:
    """
    A context manager for applying patches to functions or methods.
    This class allows you to apply patches to functions or methods within a context,
    ensuring that the original functionality is restored when exiting the context.
    """

    def __init__(
            self, 
            target: types.ModuleType | type, 
            callable_name: str, 
            prefix: PrefixFnType | None = None, 
            postfix: PostfixFnType | None = None,
            replace: ReplaceFnType | None = None,
            create: ReplaceFnType | None = None
        ):
        self.target = target
        self.callable_name = callable_name
        self.prefix = prefix
        self.postfix = postfix
        self.replace = replace
        self.create = create
    
    def __enter__(self):
        if isinstance(self.target, types.ModuleType):
            patch_function(
                self.target,
                self.callable_name,
                prefix=self.prefix,
                postfix=self.postfix,
                replace=self.replace
            )
        elif isinstance(self.target, type):
            patch_method(
                self.target,
                self.callable_name,
                prefix=self.prefix,
                postfix=self.postfix,
                replace=self.replace,
                create=self.create
            )
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if isinstance(self.target, types.ModuleType):
            unpatch_function(self.target, self.callable_name)
        elif isinstance(self.target, type):
            unpatch_method(self.target, self.callable_name)



class InjectManager:
    """
    A context manager for applying dependency injection to functions or methods.
    This class allows you to inject dependencies into functions or methods within a context,
    ensuring that the original functionality is restored when exiting the context.
    """

    def __init__(
            self,
            target: types.ModuleType | type,
            callable_name: str,
            inject_after_line: int = 0,
            code_to_inject: str | None = None
        ):
        self.target = target
        self.callable_name = callable_name
        self.inject_after_line = inject_after_line
        self.code_to_inject = code_to_inject
    
    def __enter__(self):
        if isinstance(self.target, types.ModuleType):
            inject_function(
                self.target,
                self.callable_name,
                self.inject_after_line,
                self.code_to_inject
            )
        elif isinstance(self.target, type):
            inject_method(
                self.target,
                self.callable_name,
                self.inject_after_line,
                self.code_to_inject
            )
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if isinstance(self.target, types.ModuleType):
            undo_func_inject(self.target, self.callable_name)
        elif isinstance(self.target, type):
            undo_method_inject(self.target, self.callable_name)



class HookManager:
    """
    A context manager for applying hooks to functions or methods.
    This class allows you to apply hooks to functions or methods within a context,
    ensuring that the original functionality is restored when exiting the context.
    """

    def __init__(
            self,
            target: types.ModuleType | type,
            callable_name: str,
            hook_callback: typing.Callable
        ):
        self.target = target
        self.callable_name = callable_name
        self.hook_callback = hook_callback
    
    def __enter__(self):
        if isinstance(self.target, types.ModuleType):
            register_function_hook(self.target, self.callable_name, self.hook_callback)
        elif isinstance(self.target, type):
            register_method_hook(self.target, self.callable_name, self.hook_callback)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if isinstance(self.target, types.ModuleType):
            remove_function_hook(self.target, self.callable_name, self.hook_callback)
        elif isinstance(self.target, type):
            remove_method_hook(self.target, self.callable_name, self.hook_callback)



class PatchInfo:
    """Holds informathion about a patch."""
    def __init__(self, parent, target, patch_index: int, original):
        self.parent = parent
        self.target = target
        self.patch_index = patch_index
        self.original = original



from .func_types import *
from .core import patch_method, create_method, delete_method

class Patch:
	# All patch functions
	prefix:  "PrefixFnType  | None"
	postfix: "PostfixFnType | None"
	replace: "ReplaceFnType | None"
	create:  tuple[str, "ReplaceFnType | None"]
	delete:  str | None


def apply(patch: "Patch", target_class: type, method_name: str = "__init__") -> bool:
    """
    Applies a Harmonify patch to a method of a class.
        
    Args:
        `patch`: The `Patch` that is to be applied.
        `target_class`: The class whose method is to be patched. If not provided, it defaults to "__init__".
        `method_name`: The name of the method to be patched. (as a string)
    """
    # Retrieve the patches into local variables
    patch_prefix = patch.prefix
    patch_postfix = patch.postfix
    patch_replace = patch.replace
    patch_create = patch.create
    patch_delete = patch.delete
    # Apply the main patch(es)
    patch_success = patch_method(target_class, method_name, patch_prefix, patch_postfix, patch_replace)

    create_success = True
    delete_success = True
    # Apply the creation/deletion patch(es)
    if patch_create[1]:
        create_success = create_method(target_class, patch_create[0], patch_create[1])
    if patch_delete:
        delete_success = delete_method(target_class, patch_delete)
    # Return true if all patches succeed
    return patch_success and create_success and delete_success


import functools, types



class HarmonifyDecorator:
    """
    Base Harmonify decorator class.<br>
    Extended by `no_inject` and `allow_inject`.
    """
    def __init__(self, func):
        self.func = func
        # Copy metadata from the original function
        functools.update_wrapper(self, func)

    def __call__(self, *args, **kwds):
        return self.func(*args, **kwds)

    # Implement _get_ for methods (when the decorator is applied to a method)
    # This allows the decorated method to be bound correctly to its instance.
    def __get__(self, instance, owner):
        if instance is None:
            return self # Accessing the decorator directly from the class
        
        # Return a bound method that refers to our wrapped function
        return types.MethodType(self, instance)



class no_inject(HarmonifyDecorator):
    """
    Class-based decorator to mark a function or method as not allowing Harmonify injection.
    """
    def __init__(self, func):
        super().__init__(func)

        

class allow_inject(HarmonifyDecorator):
    """
    Class-based decorator to mark a function or method as allowing Harmonify injection.
    """
    def __init__(self, func):
        super().__init__(func)




import ast
import inspect
import textwrap
import types
from . import injector_security as sec

_func_injections = {}
_method_injections = {}


def inject_function(
    target_module: types.ModuleType,
    function_name: str,
    insert_after_line: int = 0,
    code_to_inject: str | None = None
):
    """
    Inject the specified code snippet in the targeted function's source (at runtime).<br>
    If the function has a "no_inject" attribute set to True, then this will not do anything.<br>
    **Note**: This is very dangerous and allows programmmers to run *unsandboxed code*!
    
    Args:
        `target_module`: The module in which the targeted function exists.
        `function_name`: The name of the targeted function.
        `insert_after_line`: The line number (relative to the function definition) after which the code will be injected.
        `code_to_inject`: The code snippet that will be injected.
    """
    target_function = getattr(target_module, function_name)
    if isinstance(target_function, sec.no_inject):
        return False
    
    if isinstance(target_function, sec.allow_inject):
        target_function = target_function.func
    
    function_source = textwrap.dedent(inspect.getsource(target_function))
    function_ast = ast.parse(function_source)

    class CodeInjector(ast.NodeTransformer):
        def visit_FunctionDef(self, node):
            new_node = self.generic_visit(node)

            if code_to_inject:
                target_line = insert_after_line + 1
                insert_index = 0

                # Find starting place based on the line number
                for index, statement in enumerate(node.body):
                    if hasattr(statement, "lineno") and statement.lineno <= target_line:
                        insert_index = index + 1
                
                # Inject the code snippet
                injected_code = ast.parse(code_to_inject).body
                new_node.body[insert_index:insert_index] = injected_code
            
            return new_node
    
    # Transform the function's AST
    injector = CodeInjector()
    new_ast = injector.visit(function_ast)
    ast.fix_missing_locations(new_ast)   # Fix the AST's line numbers and positions

    # Compile the new function and rebind it
    compiled_func = compile(new_ast, filename=target_module.__name__, mode="exec")
    namespace = target_function.__globals__.copy()
    exec(compiled_func, namespace)

    new_function = namespace[function_name]

    # Keep track of the injection
    inject_key = (target_module, function_name)
    inject_value = (target_function, new_function)
    if inject_key not in _func_injections:
        _func_injections[inject_key] = inject_value

    # Replace the function in its original spot
    setattr(target_module, function_name, new_function)
    return True



def undo_func_inject(
    target_module: types.ModuleType,
    function_name: str
):
    """
    Revert the injected function back to its original code.

    Args:
        `target_module`: The module where the targeted function exists.
        `function_name`: The name of the targeted function.
    """
    inject_key = (target_module, function_name)
    if inject_key in _func_injections:
        original_function = _func_injections[inject_key][0]
        setattr(target_module, function_name, original_function)
    return True



def inject_method(
    target_class: type,
    method_name: str = "__init__",
    insert_after_line: int = 0,
    code_to_inject: str | None = None
):
    """
    Inject the specified code snippet in the targeted method's source (at runtime).<br>
    If the function has a "no_inject" attribute set to True, then this will not do anything.<br>
    **Note**: This is very dangerous and allows programmmers to run *unsandboxed code*!
    
    Args:
        `target_module`: The module in which the targeted method exists.
        `method_name`: The name of the targeted method.
        `insert_after_line`: The line number (relative to the method definition) after which the code will be injected.
        `code_to_inject`: The code snippet that will be injected.
    """
    target_method = getattr(target_class, method_name)
    if isinstance(target_method, sec.no_inject):
        return False
    
    if isinstance(target_method, sec.allow_inject):
        target_method = target_method.func

    # Handle method wrappers
    is_classmethod = isinstance(inspect.getattr_static(target_class, method_name), classmethod)
    is_staticmethod = isinstance(inspect.getattr_static(target_class, method_name), staticmethod)


    method_source = textwrap.dedent(inspect.getsource(target_method))
    method_ast = ast.parse(method_source)

    class CodeInjector(ast.NodeTransformer):
        def visit_FunctionDef(self, node):
            new_node = self.generic_visit(node)

            if code_to_inject:
                target_line = insert_after_line + 1
                insert_index = 0

                # Find starting place based on the line number
                for index, statement in enumerate(node.body):
                    if hasattr(statement, "lineno") and statement.lineno <= target_line:
                        insert_index = index + 1
                
                # Inject the code snippet
                injected_code = ast.parse(code_to_inject).body
                new_node.body[insert_index:insert_index] = injected_code
            
            return new_node
    
    # Transform the method's AST
    injector = CodeInjector()
    new_ast = injector.visit(method_ast)
    ast.fix_missing_locations(new_ast)   # Fix the AST's line numbers and positions

    # Compile the new method and rebind it
    compiled_func = compile(new_ast, filename=target_class.__module__, mode="exec")
    namespace = target_method.__globals__.copy()
    exec(compiled_func, namespace)

    new_method = namespace[method_name]

    # Re-wrap if needed
    if is_classmethod: new_method = classmethod(new_method)
    elif is_staticmethod: new_method = staticmethod(new_method)

    # Keep track of the injection
    inject_key = (target_class, method_name)
    inject_value = (target_method, new_method)
    if inject_key not in _method_injections:
        _method_injections[inject_key] = inject_value

    # Replace the method in its original spot
    setattr(target_class, method_name, new_method)
    return True



def undo_method_inject(
    target_class: type,
    method_name: str
):
    """
    Revert the injected method back to its original code.

    Args:
        `target_class`: The class where the targeted method exists.
        `method_name`: The name of the targeted method.
    """
    inject_key = (target_class, method_name)
    if inject_key in _method_injections:
        original_method = _method_injections[inject_key][0]
        setattr(target_class, method_name, original_method)
    return True





from .flow_control import CONTINUE_EXEC, CONTINUE_WITHOUT_POSTFIX, STOP_EXEC
from .func_types import *
import sys, inspect, types
from .info_utils import *



_method_patches = {}   # To store original methods and their patches
_function_patches = {}

n = 0
def new_id():
	global n
	n += 1
	return n


def patch_method(
	target_class: type, 
	method_name: str = "__init__", 
	prefix:  "PrefixFnType  | None" = None, 
	postfix: "PostfixFnType | None" = None, 
	replace: "ReplaceFnType | None" = None
) -> bool:
	"""
	Patches a method of a class.

	Args:
		`target_class`: The class whose method is to be patched.
		`method_name`: The name of the method to be patched (as a string). If not provided, it defaults to "__init__".
		`prefix`: A function to run *before* the original method. (optional)
		`postfix`: A function to run *after* the original method. (optional)
		`replace`: A function to completely *replace* the original method. (optional)
	"""
	original_method = getattr(target_class, method_name, None)
	if not callable(original_method):
		return False

	# Store the original method so we can restore it later
	# We'll use a unique key for each patched method
	patch_key = (target_class, method_name, new_id())
	if patch_key not in _method_patches:
		_method_patches[patch_key] = original_method

	def patched_method(instance, *args, **kwds):
		flow_state = CONTINUE_EXEC   # Assume that we're continuing the execution
		result = None

		new_args, new_kwds = args, kwds

		# This is the new method that will replace the original
		# It will handle calling prefix, original, and postfix

		# If 'replace' function is provided, it completely takes over
		if replace:
			# We need to bind 'replace' to the instance so it acts like a method
			# This makes 'self' work inside the replacement function
			bound_replace = types.MethodType(replace, instance)
			return bound_replace(*args, **kwds)

		# Call the prefix function if it exists
		if prefix:
			# Pass instance, original method, and arguments to prefix
			# Prefix can modify arguments or even return a result to skip the original or both original and postfix.
			bound_prefix = types.MethodType(prefix, instance)
			result, new_args, new_kwds, flow_state = bound_prefix(*args, **kwds)

		if flow_state != STOP_EXEC:
			# Call the original method
			# We use the stored original_method
			result = types.MethodType(original_method, instance)(*new_args, **new_kwds)

		# Call the postfix function if it exists
		if postfix and flow_state == CONTINUE_EXEC:
			# Pass instance, original method, and result to postfix
			# Postfix can modify the result
			bound_postfix = types.MethodType(postfix, instance)
			modified_result = bound_postfix(result, *args, **kwds)
			return modified_result
			
		return result

	# Replace the original method on the class
	setattr(target_class, method_name, patched_method)
	return True
	


def patch_function(
	target_module: types.ModuleType,
	function_name: str,
	prefix:  "PrefixFnType  | None" = None,
	postfix: "PostfixFnType | None" = None,
	replace: "ReplaceFnType | None" = None
) -> bool:
	"""
	Patches a standalone function in a module.

	Args:
		`target_module`: The module object where the function is defined.
		`function_name`: The name of the function to patch.
		`prefix`: A function to run *before* the original function.
		`postfix`: A function to run *after* the original function.
		`replace`: A function to completely *replace* the original function.
	"""
	original_function = getattr(target_module, function_name, None)
	if not callable(original_function):
		return False

	patch_key = (target_module, function_name, new_id())
	if patch_key not in _function_patches:
		_function_patches[patch_key] = original_function

	def patched_function(*args, **kwds):
		flow_state = CONTINUE_EXEC
		result = None

		if replace:
			return replace(*args, **kwds)
		# Call the prefix
		if prefix:
			result, new_args, new_kwds, flow_state = prefix(*args, **kwds)
		# Call the original
		if flow_state != STOP_EXEC:
			result = original_function(*new_args, **new_kwds)
		# Call the postfix
		if postfix and flow_state == CONTINUE_EXEC:
			modified_result = postfix(result, *args, **kwds)
			return modified_result

		return result

	setattr(target_module, function_name, patched_function)
	return True



def create_method(target_class: type, method_name: str, body: "ReplaceFnType") -> bool:
	"""
	Creates a new method on a class.
	
	Args:
		`target_class`: The class that the method is being added on.
		`method_name`: The name of the method that is being added.
		`body`: The body of the method.
	"""
	# No need to actually do any bounding to classes (not that we have anything to bound *to* XD)
	# A normal setattr() works too!
	setattr(target_class, method_name, body)
	return True
	


def delete_method(target_class: type, method_name: str) -> bool:
	"""
	Deletes a method on a class.
	
	Args:
		`target_class`: The class that the method is being deleted from.
		`method_name`: The name of the method that is being deleted.
	"""
	# deleting is as simple as a delattr() call
	delattr(target_class, method_name)
	return True



def delete_function(target_module: types.ModuleType, function_name: str) -> bool:
	"""
	Deletes a function from a module.

	Args:
		`target_module`: The module that the function is being deleted from.
		`function_name`: The name of the function that is being deleted.
	"""
	# deleting is as simple as a delattr() call
	delattr(target_module, function_name)
	return True



def unpatch_method(target_class: type, method_name: str = "__init__", index: int = 1) -> bool:
	"""
	Restores a patched method to its original state.
	"""
	patch_key = (target_class, method_name, index)
	if patch_key in _method_patches:
		original_method = _method_patches.pop(patch_key)
		setattr(target_class, method_name, original_method)
		return True
	return False



def unpatch_function(target_module: types.ModuleType, method_name: str, index: int = 1) -> bool:
	"""
	Restores a patched method to its original state.
	"""
	patch_key = (target_module, method_name, index)
	if patch_key in _function_patches:
		original_method = _function_patches.pop(patch_key)
		setattr(target_module, method_name, original_method)
		return True
	return False



def get_current_module() -> types.ModuleType | None:
    """
    Returns the module object of the immediate caller (i.e., the module from which this function is called).<br>
    Returns `None` if not possible.
    """
    frame = inspect.currentframe()
    if frame is None:
        return None

    try:
        caller_frame = frame.f_back
        module_name = caller_frame.f_globals.get("__name__")
        if module_name:
            return sys.modules.get(module_name)
        return None
    finally:
        # Clean up frame references to avoid reference cycles
        del frame
        del caller_frame



def get_function_patches() -> list[PatchInfo]:
	"""
	Returns a list containing all applied patches for functions.<br>
	Returns an empty list (i.e. `[]`) if no patches have been applied yet.
	"""
	patches = []
	for patch_key, original in _function_patches.items():
		(target_mod, func_name, patch_idx) = patch_key
		patches.append(PatchInfo(target_mod, func_name, patch_idx, original))
	return patches



def get_method_patches() -> list[PatchInfo]:
	"""
	Returns a list containing all applied patches for methods.<br>
	Returns an empty list (i.e. `[]`) if no patches have been applied yet.
	"""
	patches = []
	for patch_key, original in _method_patches.items():
		(target_cls, method_name, patch_idx) = patch_key
		patches.append(PatchInfo(target_cls, method_name, patch_idx, original))
	return patches



import types, typing, inspect, sys


_active_function_hooks = {}
_active_method_hooks = {}


def register_function_hook(target_module: types.ModuleType, function_name: str, hook_callback: typing.Callable):
    """
    Registers a hook callback for a function in a module.

    Args:
        `target_module`: The module object where the function is defined.
        `function_name`: The name of the function to hook.
        `hook_callback`: The callback function to execute when the hook is triggered.
    """
    hook_key = (target_module, function_name)
    if hook_key not in _active_function_hooks:
        _active_function_hooks[hook_key] = []
    _active_function_hooks[hook_key].append(hook_callback)
    return True



def register_method_hook(target_class: type, method_name: str, hook_callback: typing.Callable):
    """
    Registers a hook callback for a method in a class.

    Args:
        `target_class`: The class whose method is to be hooked.
        `method_name`: The name of the method to hook.
        `hook_callback`: The callback function to execute when the hook is triggered.
    """
    hook_key = (target_class, method_name)
    if hook_key not in _active_method_hooks:
        _active_method_hooks[hook_key] = []
    _active_method_hooks[hook_key].append(hook_callback)
    return True



def remove_function_hook(target_module: types.ModuleType, function_name: str, hook_callback: typing.Callable):
    """
    Removes a hook callback for a function in a module.

    Args:
        `target_module`: The module object where the function is defined.
        `function_name`: The name of the function to unhook.
        `hook_callback`: The callback function to remove.
    """
    hook_key = (target_module, function_name)
    if hook_key in _active_function_hooks:
        _active_function_hooks[hook_key].remove(hook_callback)
        if not _active_function_hooks[hook_key]:
            del _active_function_hooks[hook_key]
    return True



def remove_method_hook(target_class: type, method_name: str, hook_callback: typing.Callable):
    """
    Removes a hook callback for a method in a class.

    Args:
        `target_class`: The class whose method is to be unhooked.
        `method_name`: The name of the method to unhook.
        `hook_callback`: The callback function to remove.
    """
    hook_key = (target_class, method_name)
    if hook_key in _active_method_hooks:
        _active_method_hooks[hook_key].remove(hook_callback)
        if not _active_method_hooks[hook_key]:
            del _active_method_hooks[hook_key]
    return True



### THESE FUNCTIONS ARE TO BE USED IN THE LIBRARY FUNCTIONS THEMSELVES ###

def call_function_hook(hook_index: int = 0):
    """
    WARNING: This function is to be used as an API in the target library.
    Calls the function hook at the specified index.

    Args:
        `hook_index`: The index of the hook to call.
    """
    # Get the calling frame (that of the function being hooked)
    frame = inspect.currentframe().f_back
    if frame is None:
        raise RuntimeError("No calling frame found.")
    
    # Get the module and function name from the frame
    module_name = frame.f_globals.get("__name__")
    function_name = frame.f_code.co_name
    target_module = sys.modules.get(module_name)
        
    if not target_module:
        raise RuntimeError(f"Module '{module_name}' not found.")
        
    hook_key = (target_module, function_name)
    # Call the registered hook with the specified index
    if hook_key in _active_function_hooks:
        hooks = _active_function_hooks[hook_key]
        if 0 <= hook_index < len(hooks):
            hooks[hook_index]()


def call_method_hook(hook_index: int = 0):
    """
    WARNING: This function is to be used as an API in the target library.
    Calls the method hook at the specified index.
    Args:
        `hook_index`: The index of the hook to call.
    """
    # Get the calling frame (that of the function being hooked)
    frame = inspect.currentframe().f_back
    if frame is None:
        raise RuntimeError("No calling frame found.")
    
    # Get the class and method name from the frame
    class_name = frame.f_locals.get("self", None).__class__
    method_name = frame.f_code.co_name
    target_module = sys.modules.get(class_name.__module__)
        
    if not target_module:
        raise RuntimeError(f"Module '{target_module}' not found.")

    hook_key = (class_name, method_name)
    # Call the registered hook with the specified index
    if hook_key in _active_method_hooks:
        hooks = _active_method_hooks[hook_key]
        if 0 <= hook_index < len(hooks):
            hooks[hook_index]()



def call_hook(hook_index: int = 0):
    """
    Calls the hook at the specified index.
    This function is a wrapper to call either a function or method hook based on the context.
    Args:
        `hook_index`: The index of the hook to call.
    """
    # Determine if we are in a function or method context
    frame = inspect.currentframe().f_back
    if frame is None:
        raise RuntimeError("No calling frame found.")
    
    if "self" in frame.f_locals:
        call_method_hook(hook_index)
    else:
        call_function_hook(hook_index)



def get_active_function_hooks() -> dict:
    """
    Returns a dictionary of currently active function hooks.
    """
    return _active_function_hooks



def get_active_method_hooks() -> dict:
    """
    Returnst a dictionary of currently active method hooks.
    """
    return _active_method_hooks
    